""" Methods for using COBRApy to execute SED tasks in COMBINE archives and save their outputs

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-31
:Copyright: 2020, BioSimulators Team
:License: MIT
"""

from .data_model import KISAO_ALGORITHMS_PARAMETERS_MAP
from .utils import (get_active_objective_sbml_fbc_id, set_simulation_method_arg,
                    apply_variables_to_simulation_method_args, validate_variables, get_results_of_variables)
from biosimulators_utils.combine.exec import exec_sedml_docs_in_archive
from biosimulators_utils.log.data_model import CombineArchiveLog, TaskLog  # noqa: F401
from biosimulators_utils.plot.data_model import PlotFormat  # noqa: F401
from biosimulators_utils.report.data_model import ReportFormat, VariableResults  # noqa: F401
from biosimulators_utils.sedml.data_model import (Task, ModelLanguage, SteadyStateSimulation,  # noqa: F401
                                                  Variable)
from biosimulators_utils.sedml import validation
from biosimulators_utils.sedml.exec import exec_sed_doc
from biosimulators_utils.xml.utils import get_namespaces_for_xml_doc
from lxml import etree
import cobra.io
import functools

__all__ = [
    'exec_sedml_docs_in_combine_archive',
    'exec_sed_task',
]


def exec_sedml_docs_in_combine_archive(archive_filename, out_dir,
                                       report_formats=None, plot_formats=None,
                                       bundle_outputs=None, keep_individual_outputs=None):
    """ Execute the SED tasks defined in a COMBINE/OMEX archive and save the outputs

    Args:
        archive_filename (:obj:`str`): path to COMBINE/OMEX archive
        out_dir (:obj:`str`): path to store the outputs of the archive

            * CSV: directory in which to save outputs to files
              ``{ out_dir }/{ relative-path-to-SED-ML-file-within-archive }/{ report.id }.csv``
            * HDF5: directory in which to save a single HDF5 file (``{ out_dir }/reports.h5``),
              with reports at keys ``{ relative-path-to-SED-ML-file-within-archive }/{ report.id }`` within the HDF5 file

        report_formats (:obj:`list` of :obj:`ReportFormat`, optional): report format (e.g., csv or h5)
        plot_formats (:obj:`list` of :obj:`PlotFormat`, optional): report format (e.g., pdf)
        bundle_outputs (:obj:`bool`, optional): if :obj:`True`, bundle outputs into archives for reports and plots
        keep_individual_outputs (:obj:`bool`, optional): if :obj:`True`, keep individual output files

    Returns:
        :obj:`CombineArchiveLog`: log
    """
    sed_doc_executer = functools.partial(exec_sed_doc, exec_sed_task)
    return exec_sedml_docs_in_archive(sed_doc_executer, archive_filename, out_dir,
                                      apply_xml_model_changes=True,
                                      report_formats=report_formats,
                                      plot_formats=plot_formats,
                                      bundle_outputs=bundle_outputs,
                                      keep_individual_outputs=keep_individual_outputs)


def exec_sed_task(task, variables, log=None):
    ''' Execute a task and save its results

    Args:
       task (:obj:`Task`): task
       variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
       log (:obj:`TaskLog`, optional): log for the task

    Returns:
        :obj:`tuple`:

            :obj:`VariableResults`: results of variables
            :obj:`TaskLog`: log

    Raises:
        :obj:`ValueError`: if the task or an aspect of the task is not valid, or the requested output variables
            could not be recorded
        :obj:`NotImplementedError`: if the task is not of a supported type or involves an unsuported feature
    '''
    log = log or TaskLog()

    validation.validate_task(task)
    validation.validate_model_language(task.model.language, ModelLanguage.SBML)
    validation.validate_model_change_types(task.model.changes, ())
    validation.validate_model_changes(task.model.changes)
    validation.validate_simulation_type(task.simulation, (SteadyStateSimulation, ))
    validation.validate_uniform_time_course_simulation(task.simulation)
    validation.validate_data_generator_variables(variables)
    target_x_paths_ids = validation.validate_variable_xpaths(
        variables, task.model.source, attr='id')
    namespaces = get_namespaces_for_xml_doc(etree.parse(task.model.source))
    target_x_paths_fbc_ids = validation.validate_variable_xpaths(
        variables,
        task.model.source,
        attr={
            'namespace': {
                'prefix': 'fbc',
                'uri': namespaces['fbc'],
            },
            'name': 'id',
        }
    )

    # Read the model
    model = cobra.io.read_sbml_model(task.model.source)

    # get the SBML-FBC id of the active objective
    active_objective_fbc_id = get_active_objective_sbml_fbc_id(task.model.source)

    # Load the simulation method specified by ``simulation.algorithm``
    simulation = task.simulation
    algorithm_kisao_id = simulation.algorithm.kisao_id
    method_props = KISAO_ALGORITHMS_PARAMETERS_MAP.get(algorithm_kisao_id, None)
    if method_props is None:
        msg = "".join([
            "Algorithm with KiSAO id `{}` is not supported. ".format(algorithm_kisao_id),
            "Algorithm must have one of the following KiSAO ids:\n  - {}".format('\n  - '.join(
                '{}: {}'.format(kisao_id, method_props['name'])
                for kisao_id, method_props in KISAO_ALGORITHMS_PARAMETERS_MAP.items())),
        ])
        raise NotImplementedError(msg)

    # set up method parameters specified by ``simulation.algorithm.changes``
    method_kw_args = {}
    for method_arg_change in simulation.algorithm.changes:
        set_simulation_method_arg(method_props, method_arg_change, model, method_kw_args)

    # validate variables
    validate_variables(method_props, variables)

    # encode variables into arguments of the simulation metods
    apply_variables_to_simulation_method_args(target_x_paths_ids, method_props, variables, method_kw_args)

    # execute simulation
    solution = method_props['method'](model, **method_kw_args)

    # check that solution was optimal
    if method_props['check_status'] and solution.status != 'optimal':
        raise cobra.exceptions.OptimizationError("A solution could not be found. The solver status was `{}`.".format(
            solution.status))

    if method_props['kisao_id'] in ['KISAO_0000527', 'KISAO_0000528']:
        solution.objective_value = model.optimize().objective_value

    # Get the results of each variable
    variable_results = get_results_of_variables(target_x_paths_ids, target_x_paths_fbc_ids,
                                                active_objective_fbc_id, method_props, variables, solution)

    # log action
    log.algorithm = algorithm_kisao_id
    log.simulator_details = {
        'method': method_props['method'].__module__ + '.' + method_props['method'].__name__,
        'arguments': method_kw_args,
    }

    # Return the results of each variable and log
    return variable_results, log



""" Data model for mapping KiSAO terms for algorithms and their parameters
to COBRApy methods and their arguments

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-31
:Copyright: 2020, BioSimulators Team
:License: MIT
"""

from biosimulators_utils.data_model import ValueType
from numpy import nan
import cobra.flux_analysis
import enum

__all__ = ['Solver', 'KISAO_ALGORITHMS_PARAMETERS_MAP']


class Solver(str, enum.Enum):
    """ Solver """
    cplex = 'cplex'
    glpk = 'glpk'
    gurobi = 'gurobi'


KISAO_ALGORITHMS_PARAMETERS_MAP = {
    'KISAO_0000437': {
        'kisao_id': 'KISAO_0000437',
        'name': 'flux-balance analysis (FBA)',
        'method': lambda model, **args: model.optimize(**args),
        'parameters': {
            'KISAO_0000553':  {
                'name': 'solver',
                'description': 'Convex optimization solver to use (e.g., CPLEX, GLPK).',
                'model_arg': 'solver',  # "CPLEX", "GLPK", "Gurobi"
                'type': ValueType.string,
                'enum': Solver,
            }
        },
        'check_status': True,
        'variables': [
            {
                'description': 'objective value',
                'target_type': 'objective',
                'target': r'^/sbml:sbml/sbml:model/fbc:listOfObjectives/fbc:objective(\[.*?\])?(/@value)?$',
                'get_result':
                    lambda active_obj_fbc_id, el_id, el_fbc_id, solution:
                        solution.objective_value if el_fbc_id == active_obj_fbc_id else nan,
            },
            {
                'description': 'reaction flux',
                'target_type': 'reaction',
                'target': r'^/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction(\[.*?\])?(/@flux)?$',
                'get_result':
                    lambda active_obj_fbc_id, el_id, el_fbc_id, solution:
                        solution.fluxes.get(el_id[2:] if el_id.startswith('R_') else el_id),
            },
            {
                'description': 'reaction reduced cost',
                'target_type': 'reaction',
                'target': r'^/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction(\[.*?\])?/@reducedCost$',
                'get_result':
                    lambda active_obj_fbc_id, el_id, el_fbc_id, solution:
                        solution.reduced_costs.get(el_id[2:] if el_id.startswith('R_') else el_id),
            },
            {
                'description': 'species shadow price',
                'target_type': 'species',
                'target': r'^/sbml:sbml/sbml:model/sbml:listOfSpecies/sbml:species(\[.*?\])?(/@shadowPrice)?$',
                'get_result':
                    lambda active_obj_fbc_id, el_id, el_fbc_id, solution:
                        solution.shadow_prices.get(el_id[2:] if el_id.startswith('M_') else el_id),
            },
        ]
    },
    'KISAO_0000528': {
        'kisao_id': 'KISAO_0000528',
        'name': 'parsimonious flux-balance analysis (pFBA)',
        'method': cobra.flux_analysis.pfba,
        'parameters': {
            'KISAO_0000531': {
                'name': 'fraction of optimum',
                'description': 'Lower bound on the objective value relative to the optimal FBA solution.',
                'alg_arg': 'fraction_of_optimum',
                'type': ValueType.float,
            },
            'KISAO_0000553': {
                'name': 'solver',
                'description': 'Convex optimization solver to use (e.g., CPLEX, GLPK).',
                'model_arg': 'solver',
                'type': ValueType.string,
                'enum': Solver,
            },
        },
        'check_status': True,
        'variables': [
            {
                'description': 'objective value',
                'target_type': 'objective',
                'target': r'^/sbml:sbml/sbml:model/fbc:listOfObjectives/fbc:objective(\[.*?\])?(/@value)?$',
                'get_result':
                    lambda active_obj_fbc_id, el_id, el_fbc_id, solution:
                        solution.objective_value if el_fbc_id == active_obj_fbc_id else nan,
            },
            {
                'description': 'reaction flux',
                'target_type': 'reaction',
                'target': r'^/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction(\[.*?\])?(/@flux)?$',
                'get_result':
                    lambda active_obj_fbc_id, el_id, el_fbc_id, solution:
                        solution.fluxes.get(el_id[2:] if el_id.startswith('R_') else el_id),
            },
            {
                'description': 'reaction reduced cost',
                'target_type': 'reaction',
                'target': r'^/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction(\[.*?\])?/@reducedCost$',
                'get_result':
                    lambda active_obj_fbc_id, el_id, el_fbc_id, solution:
                        solution.reduced_costs.get(el_id[2:] if el_id.startswith('R_') else el_id),
            },
            {
                'description': 'species shadow price',
                'target_type': 'species',
                'target': r'^/sbml:sbml/sbml:model/sbml:listOfSpecies/sbml:species(\[.*?\])?(/@shadowPrice)?$',
                'get_result':
                    lambda active_obj_fbc_id, el_id, el_fbc_id, solution:
                        solution.shadow_prices.get(el_id[2:] if el_id.startswith('M_') else el_id),
            },
        ]
    },
    'KISAO_0000527': {
        'kisao_id': 'KISAO_0000527',
        'name': 'geometric flux-balance analysis (gFBA)',
        'method': cobra.flux_analysis.geometric_fba,
        'parameters': {
            'KISAO_0000209': {
                'name': 'epsilon',
                'description': 'Convergence tolerance.',
                'alg_arg': 'epsilon',
                'type': ValueType.float,
            },
            'KISAO_0000486': {
                'name': 'max tries',
                'description': 'Maximum number of iterations.',
                'alg_arg': 'max_tries',
                'type': ValueType.integer,
            },
            'KISAO_0000529': {
                'name': 'processes',
                'description': 'Number of parallel processes to execute.',
                'alg_arg': 'processes',
                'type': ValueType.integer,
            },
            'KISAO_0000553': {
                'name': 'solver',
                'description': 'Convex optimization solver to use (e.g., CPLEX, GLPK).',
                'model_arg': 'solver',
                'type': ValueType.string,
                'enum': Solver,
            },
        },
        'check_status': True,
        'variables': [
            {
                'description': 'objective value',
                'target_type': 'objective',
                'target': r'^/sbml:sbml/sbml:model/fbc:listOfObjectives/fbc:objective(\[.*?\])?(/@value)?$',
                'get_result':
                    lambda active_obj_fbc_id, el_id, el_fbc_id, solution:
                        solution.objective_value if el_fbc_id == active_obj_fbc_id else nan,
            },
            {
                'description': 'reaction flux',
                'target_type': 'reaction',
                'target': r'^/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction(\[.*?\])?(/@flux)?$',
                'get_result':
                    lambda active_obj_fbc_id, el_id, el_fbc_id, solution:
                        solution.fluxes.get(el_id[2:] if el_id.startswith('R_') else el_id),
            },
            {
                'description': 'reaction reduced cost',
                'target_type': 'reaction',
                'target': r'^/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction(\[.*?\])?/@reducedCost$',
                'get_result':
                    lambda active_obj_fbc_id, el_id, el_fbc_id, solution:
                        nan,
            },
            {
                'description': 'species shadow price',
                'target_type': 'species',
                'target': r'^/sbml:sbml/sbml:model/sbml:listOfSpecies/sbml:species(\[.*?\])?(/@shadowPrice)?$',
                'get_result':
                    lambda active_obj_fbc_id, el_id, el_fbc_id, solution:
                        nan,
            },
        ]
    },
    'KISAO_0000526': {
        'kisao_id': 'KISAO_0000526',
        'name': 'flux variability analysis (FVA)',
        'method': cobra.flux_analysis.flux_variability_analysis,
        'parameters': {
            'KISAO_0000532': {
                'name': 'loopless',
                'description': 'Whether to return only loopless solutions.',
                'alg_arg': 'loopless',
                'type': ValueType.boolean,
            },
            'KISAO_0000531': {
                'name': 'fraction of optimum',
                'description': 'Lower bound on the objective value relative to the optimal FBA solution.',
                'alg_arg': 'fraction_of_optimum',
                'type': ValueType.float,
            },
            'KISAO_0000533': {
                'name': 'pFBA factor',
                'description': 'Upper bound on the total sum of absolute fluxes relative to the smallest feasible sum of absolute fluxes.',
                'alg_arg': 'pfba_factor',
                'type': ValueType.float,
            },
            'KISAO_0000529': {
                'name': 'processes',
                'description': 'Number of parallel processes to execute.',
                'alg_arg': 'processes',
                'type': ValueType.integer,
            },
            'KISAO_0000553': {
                'name': 'solver',
                'description': 'Convex optimization solver to use (e.g., CPLEX, GLPK).',
                'model_arg': 'solver',
                'type': ValueType.string,
                'enum': Solver,
            },
        },
        'check_status': False,
        'variables': [
            {
                'description': 'reaction flux',
                'target_type': 'reaction',
                'target': r'^/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction(\[.*?\])?/@minFlux?$',
                'get_result':
                    lambda active_obj_fbc_id, el_id, el_fbc_id, solution:
                        solution.loc[el_id[2:] if el_id.startswith('R_') else el_id, 'minimum'],
            },
            {
                'description': 'reaction flux',
                'target_type': 'reaction',
                'target': r'^/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction(\[.*?\])?/@maxFlux?$',
                'get_result':
                    lambda active_obj_fbc_id, el_id, el_fbc_id, solution:
                        solution.loc[el_id[2:] if el_id.startswith('R_') else el_id, 'maximum'],
            },
        ],
    },
}



""" BioSimulators-compliant command-line interface to the `COBRApy <https://opencobra.github.io/cobrapy>`_ simulation program.

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-31
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ._version import __version__
from .core import exec_sedml_docs_in_combine_archive
from biosimulators_utils.simulator.cli import build_cli
import cobra

App = build_cli('cobrapy', __version__,
                'COBRApy', cobra.__version__, 'https://opencobra.github.io/cobrapy',
                exec_sedml_docs_in_combine_archive)


def main():
    with App() as app:
        app.run()



""" Utilities for working with COBRApy

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-31
:Copyright: 2020, BioSimulators Team
:License: MIT
"""

from biosimulators_utils.report.data_model import VariableResults
from biosimulators_utils.sedml.data_model import Variable  # noqa: F401
from biosimulators_utils.utils.core import validate_str_value, parse_value
import cobra  # noqa: F401
import libsbml
import numpy
import re

__all__ = [
    'get_active_objective_sbml_fbc_id',
    'set_simulation_method_arg',
    'apply_variables_to_simulation_method_args',
    'validate_variables',
    'get_results_of_variables',
]


def get_active_objective_sbml_fbc_id(model_source):
    """ Get the SBML-FBC id of the active objective

    Args:
        model_source (:obj:`str`): path to model

    Returns:
        :obj:`str`: SBML-FBC id of the active objective
    """
    doc = libsbml.readSBMLFromFile(model_source)
    model = doc.getModel()
    model_fbc = model.getPlugin("fbc")
    return model_fbc.getListOfObjectives().getActiveObjective()


def set_simulation_method_arg(method_props, argument_change, model, model_method_kw_args):
    """ Set the value of an argument of a simulation method based on a SED
    algorithm parameter change

    Args:
        method_props (:obj:`dict`): properties of the simulation method
        argument_change (:obj:`AlgorithmParameterChange`): algorithm parameter change
        model (:obj:`cobra.core.model.Model`)
        model_method_kw_args (:obj:`dict`): keyword arguments for the simulation method
            for the model

    Raises:
        :obj:`NotImplementedError`: if the simulation method doesn't support the parameter
        :obj:`ValueError`: if the new value is not a valid value of the parameter
    """
    parameter_kisao_id = argument_change.kisao_id
    parameter = method_props['parameters'].get(parameter_kisao_id, None)
    if parameter is None:
        msg = "".join([
            "{} ({}) does not support parameter `{}`. ".format(
                method_props['name'], method_props['kisao_id'], argument_change.kisao_id),
            "The parameters of {} must have one of the following KiSAO ids:\n  - {}".format(
                method_props['name'],
                '\n  - '.join(
                    '{} ({}): {}'.format(kisao_id, parameter['name'], parameter['description'])
                    for kisao_id, parameter in method_props['parameters'].items())),
        ])
        raise NotImplementedError(msg)

    value = argument_change.new_value
    if not validate_str_value(value, parameter['type']):
        msg = "`{}` is not a valid value for parameter {} ({}) of {} ({})".format(
            value, parameter['name'], parameter_kisao_id,
            method_props['name'], method_props['kisao_id'])
        raise ValueError(msg)
    enum = parameter.get('enum', None)
    if enum:
        if value.lower() not in enum.__members__:
            msg = ("`{}` is not a valid value for parameter {} ({}) of {} ({}). "
                   "The value of {} must be one of the following:\n  - {}").format(
                value, parameter['name'], parameter_kisao_id,
                method_props['name'], method_props['kisao_id'],
                method_props['name'],
                '\n  - '.join(sorted('`' + value + '`' for value in enum.__members__.keys())))
            raise ValueError(msg)

    parsed_value = parse_value(value, parameter['type'])
    if enum:
        parsed_value = enum[value.lower()].value

    if 'alg_arg' in parameter:
        model_method_kw_args[parameter['alg_arg']] = parsed_value
    else:
        setattr(model, parameter['model_arg'], parsed_value)


def apply_variables_to_simulation_method_args(target_x_paths_ids, method_props, variables, model_method_kw_args):
    """ Encode the desired output variables into arguments to simulation methods

    Args:
        target_x_paths_ids (:obj:`dict` of :obj:`str` to :obj:`str`): dictionary that maps each XPath to the
            SBML id of the corresponding model object
        method_props (:obj:`dict`): properties of the simulation method
        variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
        model_method_kw_args (:obj:`dict`): keyword arguments for the simulation method
            for the model
    """
    if method_props['kisao_id'] == 'KISAO_0000526':
        reaction_list = set()
        for variable in variables:
            reaction_id = target_x_paths_ids[variable.target]
            reaction_list.add(reaction_id[2:] if reaction_id.startswith('R_') else reaction_id)
        model_method_kw_args['reaction_list'] = sorted(reaction_list)


def validate_variables(method, variables):
    """ Validate the desired output variables of a simulation

    Args:
        method (:obj:`dict`): properties of desired simulation method
        variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
    """
    invalid_symbols = set()
    invalid_targets = set()
    for variable in variables:
        if variable.symbol:
            invalid_symbols.add(variable.symbol)

        else:
            valid = False
            for variable_pattern in method['variables']:
                if re.match(variable_pattern['target'], variable.target):
                    valid = True
                    break

            if not valid:
                invalid_targets.add(variable.target)

    if invalid_symbols:
        raise NotImplementedError("{} ({}) doesn't support variables with symbols".format(
            method['name'], method['kisao_id']))

    if invalid_targets:
        msg = (
            "{} ({}) doesn't support variables with the following target XPATHs:\n  - {}\n\n"
            "The targets of variables should match one of the following patterns of XPATHs:\n  - {}"
        ).format(
            method['name'], method['kisao_id'],
            '\n  - '.join(sorted('`' + target + '`' for target in invalid_targets)),
            '\n  - '.join(sorted('{}: `{}`'.format(
                variable_pattern['description'], variable_pattern['target'])
                for variable_pattern in method['variables']))
        )
        raise ValueError(msg)


def get_results_of_variables(target_x_paths_ids, target_x_paths_fbc_ids, active_objective_fbc_id, method, variables, solution):
    """ Get the results of the desired variables

    Args:
        target_x_paths_ids (:obj:`dict` of :obj:`str` to :obj:`str`): dictionary that maps each XPath to the
            SBML id of the corresponding model object
        target_x_paths_fbc_ids (:obj:`dict` of :obj:`str` to :obj:`str`): dictionary that maps each XPath to the
            SBML-FBC id of the corresponding model object
        active_objective_fbc_id (:obj:`str`): SBML-FBC id of the active objective
        method (:obj:`dict`): properties of desired simulation method
        variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
        solution (:obj:`cobra.core.solution.Solution`): solution of method

    Returns:
        :obj:`VariableResults`: the results of desired variables
    """
    variable_results = VariableResults()

    for variable in variables:
        target = variable.target
        for variable_pattern in method['variables']:
            if re.match(variable_pattern['target'], target):
                variable_target_id = target_x_paths_ids[target]
                variable_target_fbc_id = target_x_paths_fbc_ids[target]
                result = variable_pattern['get_result'](active_objective_fbc_id, variable_target_id,
                                                        variable_target_fbc_id, solution)

                break

        variable_results[variable.id] = numpy.array(result)

    return variable_results
